/*
 * This file is a part of PikoTools
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2023, Tomasz Sowa
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef headerfile_pikotools_src_space_keyvalueparser
#define headerfile_pikotools_src_space_keyvalueparser

#include "space.h"
#include "convert/baseparser.h"



namespace pt
{



class KeyValueParser : public BaseParser
{
public:


	/*
	 * ctor -- sets default values
	*/
	KeyValueParser();


	/*
	 * status of parsing
	 * ok - input stream has been parsed correctly
	 * cant_open_file - I cannot open the file (returns only in a case when parsing a file)
	 *
	*/
	enum Status {
		ok,
		cant_open_file,
	};


	/*
	 * the last status of parsing, set by parse() methods
	*/
	Status status;



	/*
	 * if true then the input file or string (char* or std::string) is treated as UTF-8
	 * default true
	 *
	 * the internal storage for strings is std::wstring so if you call set_utf8(false) then
	 * the characters of input string will be simple static_cast<> from char to wchar_t
	 *
	 */
	void use_utf8(bool utf);


	/*
	 * set a separator between a variable and a value
	 * default: '='
	 */
	void set_separator(wchar_t separator);


	/*
	 * set an option delimiter
	 * default: ','
	 */
	void set_option_delimiter(wchar_t option_delimiter);


	/*
	 *
	 */
	Status parse_file(const char * file_name, Space & out_space, bool clear_space = true);
	Status parse_file(const std::string & file_name, Space & out_space, bool clear_space = true);
	Status parse_file(const wchar_t * file_name, Space & out_space, bool clear_space = true);
	Status parse_file(const std::wstring & file_name, Space & out_space, bool clear_space = true);

	/*
	 *
	 */
	Status parse(const char * str, Space & out_space, bool clear_space = true);
	Status parse(const std::string & str, Space & out_space, bool clear_space = true);
	Status parse(const wchar_t * str, Space & out_space, bool clear_space = true);
	Status parse(const std::wstring & str, Space & out_space, bool clear_space = true);
	Status parse(const pt::TextStream & str, Space & out_space, bool clear_space = true);
	Status parse(const pt::WTextStream & str, Space & out_space, bool clear_space = true);



private:


	/*
	 * current output space
	 *
	 */
	Space * root_space;


	/*
	 * separator between a variable and a value, default: '='
	 */
	wchar_t separator;


	/*
	 * option delimiter, default: ','
	 */
	wchar_t option_delimiter;


	/*
	 *
	 */
	std::wstring key;


	/*
	 *
	 */
	std::wstring value;



	void prepare_to_parsing();
	void parse(bool clear_root_space);
	wchar_t read_token(std::wstring & token, wchar_t delimit1, wchar_t delimit2 = -1);
	void skip_white();
	int read_char();


};




} // namespace


#endif
